from __future__ import annotations

import mitzu.model as M
import mitzu.adapters.generic_adapter as GA
from typing import Tuple
import pandas as pd
import mitzu.visualization.common as C

STEP_COL = "_step"


def get_conversion_cols(metric: M.ConversionMetric):
    funnel_length = len(metric._conversion._segments)
    return [f"{GA.AGG_VALUE_COL}_{index}" for index in range(1, funnel_length + 1)]


def fix_conv_times_pdf(
    pdf: pd.DataFrame, metric: M.ConversionMetric
) -> Tuple[pd.DataFrame, int]:
    cols = get_conversion_cols(metric)
    max_ttc = pdf[cols].max(axis=1).max(axis=0)

    for key in cols:
        if 0 <= max_ttc <= C.TTC_RANGE_1_SEC:
            pdf[key] = pdf[key].round(1)
        elif C.TTC_RANGE_1_SEC < max_ttc <= C.TTC_RANGE_2_SEC:
            pdf[key] = pdf[key].div(60).round(1)
        elif C.TTC_RANGE_2_SEC < max_ttc <= C.TTC_RANGE_3_SEC:
            pdf[key] = pdf[key].div(3600).round(1)
        else:
            pdf[key] = pdf[key].div(86400).round(1)
    return pdf


def fix_conversion_na_cols(
    pdf: pd.DataFrame, metric: M.ConversionMetric
) -> pd.DataFrame:
    pdf[GA.GROUP_COL] = pdf[GA.GROUP_COL].fillna("n/a")
    funnel_length = len(metric._conversion._segments)
    cols = [f"{GA.AGG_VALUE_COL}_{i}" for i in range(1, funnel_length + 1)]
    pdf[cols] = pdf[cols].fillna(0)
    return pdf


def get_melted_conv_column(
    column_prefix: str, display_name: str, pdf: pd.DataFrame, metric: M.ConversionMetric
) -> pd.DataFrame:
    res = pd.melt(
        pdf,
        id_vars=[GA.GROUP_COL],
        value_vars=[
            f"{column_prefix}{i+1}" for i, _ in enumerate(metric._conversion._segments)
        ],
        var_name=STEP_COL,
        value_name=display_name,
    )
    res[STEP_COL] = res[STEP_COL].replace(
        {
            f"{column_prefix}{i+1}": f"<b>Step {i+1}.</b>"
            for i, val in enumerate(metric._conversion._segments)
        }
    )
    return res


def get_melted_conv_pdf(pdf: pd.DataFrame, metric: M.ConversionMetric) -> pd.DataFrame:
    pdf1 = get_melted_conv_column(f"{GA.AGG_VALUE_COL}_", GA.AGG_VALUE_COL, pdf, metric)
    pdf2 = get_melted_conv_column(
        f"{GA.USER_COUNT_COL}_", GA.USER_COUNT_COL, pdf, metric
    )
    res = pdf1
    res = pd.merge(
        res,
        pdf2,
        left_on=[GA.GROUP_COL, STEP_COL],
        right_on=[GA.GROUP_COL, STEP_COL],
    )
    return res


def get_conversion_value_suffix(pdf: pd.DataFrame, metric: M.ConversionMetric) -> str:
    if metric._agg_type in [
        M.AggType.AVERAGE_TIME_TO_CONV,
        M.AggType.PERCENTILE_TIME_TO_CONV,
    ]:
        cols = get_conversion_cols(metric)
        max_ttc = pdf[cols].max(axis=1).max(axis=0)
        return (
            "secs"
            if max_ttc <= C.TTC_RANGE_1_SEC
            else "mins"
            if max_ttc <= C.TTC_RANGE_2_SEC
            else "hours"
            if max_ttc <= C.TTC_RANGE_3_SEC
            else "days"
        )
    else:
        return "%"


def get_conversion_mapping(pdf: pd.DataFrame, metric: M.ConversionMetric):
    mapping = {
        GA.AGG_VALUE_COL: C.Y_AXIS_COL,
        GA.GROUP_COL: C.COLOR_COL,
    }
    if metric._time_group == M.TimeGroup.TOTAL:
        mapping[STEP_COL] = C.X_AXIS_COL
    else:
        mapping[GA.DATETIME_COL] = C.X_AXIS_COL
    return pdf.rename(columns=mapping)



import traceback
from typing import Any, Dict, List, Optional, cast, Union

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import flask
from dash import (
    ALL,
    Input,
    Output,
    State,
    callback,
    ctx,
    html,
    register_page,
    no_update,
)
from mitzu.webapp.helper import MITZU_LOCATION

import mitzu.helper as H
import mitzu.model as M
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.helper as WH
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
import mitzu.webapp.pages.projects.helper as MPH
import mitzu.webapp.pages.projects.manage_project_component as MPC

from mitzu.webapp.auth.decorator import restricted, restricted_layout
from datetime import datetime

CREATE_PROJECT_DOCS_LINK = "https://github.com/mitzu-io/mitzu/blob/main/DOCS.md"
PROJECT_TITLE = "project_title"
SAVE_BUTTON = "project_save_button"
MANAGE_PROJECT_INFO = "manage_project_info"

DELETE_BUTTON = "project_delete_button"

CONFIRM_DIALOG_INDEX = "project_delete_confirm"
CONFIRM_DIALOG_CLOSE = "project_delete_confirm_dialog_close"
CONFIRM_DIALOG_ACCEPT = "project_delete_confirm_dialog_accept"


def create_delete_button(project: Optional[M.Project]) -> bc.Component:
    if project is not None:
        return dbc.Button(
            [html.B(className="bi bi-x-circle me-1"), "Delete Project"],
            id=DELETE_BUTTON,
            color="danger",
            class_name="d-inline-block me-3 mb-1",
        )
    else:
        return html.Div()


def create_confirm_dialog(project: Optional[M.Project]):
    if project is None:
        return html.Div()
    return dbc.Modal(
        [
            dbc.ModalBody(
                f"Do you really want to delete the {project.project_name}?",
                class_name="lead",
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        "Close",
                        id=CONFIRM_DIALOG_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        "Delete",
                        id=CONFIRM_DIALOG_ACCEPT,
                        size="sm",
                        color="danger",
                        href=P.PROJECTS_PATH,
                        external_link=True,
                    ),
                ]
            ),
        ],
        id=CONFIRM_DIALOG_INDEX,
        is_open=False,
    )


def create_event_data_table(project: M.Project, tr: html.Tr):
    full_table_name = MPH.get_value_from_row(tr, 1)
    user_id_column = MPH.get_value_from_row(tr, 2)
    event_time_column = MPH.get_value_from_row(tr, 3)
    event_name_column = MPH.get_value_from_row(tr, 4)
    date_partition_col = MPH.get_value_from_row(tr, 5)
    ignore_cols = MPH.get_value_from_row(tr, 6)

    schema, table_name = tuple(full_table_name.split("."))
    adapter = project.get_adapter()

    fields = adapter.list_all_table_columns(schema, table_name)
    all_fields: Dict[str, M.Field] = {}
    for field in fields:
        if field._sub_fields:
            for sf in field.get_all_subfields():
                all_fields[sf._get_name()] = sf
        else:
            all_fields[field._get_name()] = field

    converted_ignored_fields: List[Union[str, M.Field]] = []
    if ignore_cols:
        for igf in ignore_cols.split(","):
            converted_ignored_fields.append(all_fields[igf])

    return M.EventDataTable.create(
        table_name=table_name,
        schema=schema,
        event_time_field=all_fields[event_time_column],
        event_name_field=all_fields[event_name_column] if event_name_column else None,
        ignored_fields=converted_ignored_fields,
        user_id_field=all_fields[user_id_column],
        date_partition_field=(
            all_fields[date_partition_col] if date_partition_col else None
        ),
    )


@restricted_layout
def layout_create() -> bc.Component:
    return layout(None)


@restricted_layout
def layout(project_id: Optional[str] = None, **query_params) -> bc.Component:
    project: Optional[M.Project] = None
    dependencies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    if project_id is not None:
        project = dependencies.storage.get_project(project_id)

    title = (
        "Create new project"
        if project is None
        else f"{H.value_to_label(project.project_name)}"
    )

    return html.Div(
        [
            NB.create_mitzu_navbar("create-project-navbar", []),
            dbc.Container(
                children=[
                    dbc.Row(
                        [
                            dbc.Col(
                                html.H4(
                                    title, id=PROJECT_TITLE, className="card-title"
                                ),
                                width="auto",
                            )
                        ]
                    ),
                    html.Hr(),
                    MPC.create_project_settings(project, dependencies),
                    html.Hr(),
                    dbc.Button(
                        [html.B(className="bi bi-check-circle"), " Save"],
                        color="success",
                        id=SAVE_BUTTON,
                        class_name="d-inline-block me-3 mb-1",
                    ),
                    dbc.Button(
                        [
                            html.B(className="bi bi-search me-1"),
                            (
                                "Discover Project"
                                if project_id is not None
                                else "Discover Projects"
                            ),
                        ],
                        color="primary",
                        class_name="d-inline-block me-3 mb-1",
                        href=(
                            P.create_path(
                                P.EVENTS_AND_PROPERTIES_PROJECT_PATH,
                                project_id=project_id,
                            )
                            if project_id is not None
                            else P.EVENTS_AND_PROPERTIES_PROJECT_PATH
                        ),
                    ),
                    html.Div(
                        children="",
                        className="mb-3 lead",
                        id=MANAGE_PROJECT_INFO,
                    ),
                    html.Hr(),
                    create_delete_button(project),
                    create_confirm_dialog(project),
                ],
                class_name="mb-3",
            ),
        ]
    )


@callback(
    Output(CONFIRM_DIALOG_INDEX, "is_open"),
    Input(DELETE_BUTTON, "n_clicks"),
    Input(CONFIRM_DIALOG_CLOSE, "n_clicks"),
    prevent_initial_call=True,
)
@restricted
def delete_button_clicked(delete: int, close: int) -> bool:
    if delete is None:
        return no_update
    return ctx.triggered_id == DELETE_BUTTON


@callback(
    Output(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    State(MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def delete_confirm_button_clicked(n_clicks: int, pathname: str) -> int:
    if n_clicks:
        project_id = P.get_path_value(
            P.PROJECTS_MANAGE_PATH, pathname, P.PROJECT_ID_PATH_PART
        )
        depenednecies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )
        try:
            project = depenednecies.storage.get_project(project_id=project_id)
            for edt in project.event_data_tables:
                depenednecies.storage.delete_event_data_table_definition(
                    project_id=project_id, edt_full_name=edt.get_full_name()
                )
            depenednecies.storage.delete_project(project_id)
        except Exception:
            # TBD: Toaster
            traceback.print_exc()

    return no_update


@callback(
    Output(MANAGE_PROJECT_INFO, "children"),
    Input(SAVE_BUTTON, "n_clicks"),
    State(MPH.EDT_TBL_BODY, "children"),
    State({"type": MPH.PROJECT_INDEX_TYPE, "index": ALL}, "value"),
    State(WH.MITZU_LOCATION, "pathname"),
    prevent_initial_call=True,
)
@restricted
def save_button_clicked(
    save_clicks: int, edt_table_rows: List, prop_values: List, pathname: str
):
    try:
        storage = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        ).storage

        project_props: Dict[str, Any] = {}

        for prop in ctx.args_grouping[2]:
            id_val = prop["id"]
            if id_val.get("type") == MPH.PROJECT_INDEX_TYPE:
                project_props[id_val.get("index")] = prop["value"]

        project_id = cast(str, project_props.get(MPC.PROP_PROJECT_ID))
        project_name = cast(str, project_props.get(MPC.PROP_PROJECT_NAME))
        if not project_name:
            return html.P("Please name your project first!", className="text-danger")

        connection_id = cast(str, project_props.get(MPC.PROP_CONNECTION))
        if connection_id is None:
            return html.P("Please select a connection first!", className="text-danger")

        description = cast(str, project_props.get(MPC.PROP_DESCRIPTION))
        disc_lookback_days = cast(int, project_props.get(MPC.PROP_DISC_LOOKBACK_DAYS))
        min_sample_size = cast(int, project_props.get(MPC.PROP_DISC_SAMPLE_SIZE))
        autorefresh_enabled = cast(
            bool, project_props.get(MPC.PROP_EXPLORE_AUTO_REFRESH)
        )
        end_date_config = M.WebappEndDateConfig[
            project_props.get(
                MPC.PROP_END_DATE_CONFIG, M.WebappEndDateConfig.NOW.name
            ).upper()
        ]
        custom_end_date_str = project_props.get(MPC.PROP_CUSTOM_END_DATE_CONFIG)
        if custom_end_date_str is not None:
            custom_end_date = datetime.strptime(custom_end_date_str[:10], "%Y-%m-%d")
        else:
            custom_end_date = None

        connection = storage.get_connection(connection_id)
        dummy_project = M.Project(
            connection=connection,
            event_data_tables=[],
            project_name="dummy_project",
        )

        event_data_tables = []
        for tr in edt_table_rows:
            event_data_tables.append(create_event_data_table(dummy_project, tr))

        project = M.Project(
            project_name=project_name,
            project_id=project_id,
            connection=connection,
            description=description,
            webapp_settings=M.WebappSettings(
                auto_refresh_enabled=autorefresh_enabled,
                end_date_config=end_date_config,
                custom_end_date=custom_end_date,
            ),
            discovery_settings=M.DiscoverySettings(
                min_property_sample_size=min_sample_size,
                lookback_days=disc_lookback_days,
            ),
            event_data_tables=event_data_tables,
        )

        storage.set_project(project_id, project)

        return "Project succesfully saved"
    except Exception as exc:
        traceback.print_exc()
        return f"Something went wrong: {str(exc)}"


register_page(
    __name__ + "_create",
    path=P.PROJECTS_CREATE_PATH,
    title="Mitzu - Create Project",
    layout=layout_create,
)


register_page(
    __name__,
    path_template=P.PROJECTS_MANAGE_PATH,
    title="Mitzu - Manage Project",
    layout=layout,
)



from __future__ import annotations

import dash
import os
import time
import traceback
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import flask
import werkzeug
import jwt
import requests
import base64
from typing import Any, Dict, Optional, List
from urllib import parse
from mitzu.helper import LOGGER
import mitzu.webapp.pages.paths as P
import mitzu.webapp.configs as configs
import mitzu.webapp.service.user_service as U

HOME_URL = os.getenv("HOME_URL", "http://localhost:8082")
MITZU_WEBAPP_URL = os.getenv("MITZU_WEBAPP_URL", HOME_URL)
JWT_ALGORITHM = "HS256"


@dataclass(frozen=True)
class OAuthConfig:
    """
    Contains the minimal configuration for an OAuth backend.

    :param client_id: Client ID, used for validing the JWT token claim and for fetching the identity token
    :param client_secret: Client secret, used for fetching the identity toke
    :param jwks_url: URL to fetch the JSON Web Key Set (JWKS)
    :param sign_in_url: URL to where the user is redirected at the beginning of the sign in flow
    :param sign_out_url: default is None, if set then the user is redirected to this URL at the nd of the sing out flow
    :param token_url: URL where the tokens can be fetched during the sign in flow
    :param jwt_algorithms: List of supported signing algorithms for the JWT tokens
    """

    client_id: str
    client_secret: str
    jwks_url: str
    sign_in_url: str
    sign_out_url: Optional[str]
    token_url: str
    jwt_algorithms: List[str]


class TokenValidator(ABC):
    @abstractmethod
    def validate_token(self, token: str) -> Dict[str, Any]:
        pass


class JWTTokenValidator(TokenValidator):
    def __init__(self, jwks_url: str, algorithms: List[str], audience: str):
        self._jwks_client = jwt.PyJWKClient(jwks_url)
        self._algorithms = algorithms
        self._audience = audience

    def validate_token(self, token: str) -> Dict[str, Any]:
        signing_key = self._jwks_client.get_signing_key_from_jwt(token)

        return jwt.decode(
            token,
            signing_key.key,
            algorithms=self._algorithms,
            audience=self._audience,
        )

    @staticmethod
    def create_from_oauth_config(oauth_config: OAuthConfig) -> JWTTokenValidator:
        return JWTTokenValidator(
            oauth_config.jwks_url,
            oauth_config.jwt_algorithms,
            oauth_config.client_id,
        )


@dataclass(frozen=True)
class AuthConfig:
    token_cookie_name: str = field(default_factory=lambda: "auth-token")
    redirect_cookie_name: str = field(default_factory=lambda: "redirect-to")

    session_timeout: int = field(default_factory=lambda: 7 * 24 * 60 * 60)
    token_signing_key: str = field(default_factory=lambda: configs.AUTH_JWT_SECRET)

    oauth: Optional[OAuthConfig] = None
    token_validator: Optional[TokenValidator] = None
    allowed_email_domain: Optional[str] = None

    user_service: Optional[U.UserService] = None


@dataclass(frozen=True)
class OAuthAuthorizer:

    _authorized_url_prefixes: List[str] = field(
        default_factory=lambda: [
            P.UNAUTHORIZED_URL,
            configs.HEALTH_CHECK_PATH,
            "/assets/",
            "/_dash-update-component",
            "/_dash-component-suites/",
            "/_dash-layout",
            "/_dash-dependencies",
        ]
    )
    _ignore_token_refresh_prefixes: List[str] = field(
        default_factory=lambda: [
            P.UNAUTHORIZED_URL,
            P.SIGN_OUT_URL,
            configs.HEALTH_CHECK_PATH,
            "/assets/",
        ]
    )
    _config: AuthConfig = field(default_factory=lambda: AuthConfig())

    @classmethod
    def create(cls, config: AuthConfig) -> OAuthAuthorizer:
        return OAuthAuthorizer(
            _config=config,
        )

    def _get_unauthenticated_response(
        self, redirect: Optional[str] = None
    ) -> werkzeug.wrappers.response.Response:
        resp = self._redirect(P.UNAUTHORIZED_URL)
        resp.set_cookie(self._config.token_cookie_name, "", expires=0)
        if (
            redirect
            and not redirect.startswith("/assets/")
            and not redirect.startswith("/_dash")
            and not redirect.startswith(configs.HEALTH_CHECK_PATH)
        ):
            resp.set_cookie(self._config.redirect_cookie_name, redirect)
        return resp

    def _redirect(self, location: str) -> werkzeug.wrappers.response.Response:
        resp = flask.redirect(code=307, location=location)
        resp.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        resp.headers["Pragma"] = "no-cache"
        resp.headers["Expires"] = "0"
        resp.headers["Cache-Control"] = "public, max-age=0"
        return resp

    def _get_oauth_code(self) -> Optional[str]:
        code = flask.request.values.get("code")
        if code is not None:
            return code
        parse_result = parse.urlparse(flask.request.url)
        params = parse.parse_qs(parse_result.query)
        code_ls = params.get("code")
        if code_ls is not None:
            return code_ls[0]
        return None

    def _get_identity_token(self, auth_code) -> str:
        if not self._config.oauth:
            raise ValueError("OAuth is not configured")
        message = bytes(
            f"{self._config.oauth.client_id}:{self._config.oauth.client_secret}",
            "utf-8",
        )
        secret_hash = base64.b64encode(message).decode()
        payload = {
            "grant_type": "authorization_code",
            "client_id": self._config.oauth.client_id,
            "code": auth_code,
            "redirect_uri": f"{HOME_URL}{P.OAUTH_CODE_URL}",
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Authorization": f"Basic {secret_hash}",
        }

        resp = requests.post(
            self._config.oauth.token_url, params=payload, headers=headers
        )

        if resp.status_code != 200:
            raise Exception(
                f"Unexpected response: {resp.status_code}, {resp.content.decode('utf-8')}"
            )

        return resp.json()["id_token"]

    def _generate_new_token_for_identity(
        self, identity: str, role: U.Role = U.Role.MEMBER
    ) -> str:
        now = int(time.time())
        claims = {
            "iat": now - 10,
            "exp": now + self._config.session_timeout,
            "iss": "mitzu",
            "sub": identity,
            "rol": role.value,
        }
        return jwt.encode(
            claims, key=self._config.token_signing_key, algorithm=JWT_ALGORITHM
        )

    def _validate_token(self, token: str) -> Optional[Dict]:
        try:
            claims = jwt.decode(
                token, self._config.token_signing_key, algorithms=[JWT_ALGORITHM]
            )
            if "rol" in claims.keys():
                claims["rol"] = U.Role(claims["rol"])
            else:
                claims[
                    "rol"
                ] = U.Role.MEMBER  # backward compatibility with old sessions
            return claims
        except Exception as e:
            LOGGER.warning(f"Failed to validate token: {str(e)}")
            return None

    def _validate_foreign_token(self, token) -> Optional[str]:
        if not self._config.token_validator:
            raise ValueError("Token validator is not configured")
        try:
            decoded_token = self._config.token_validator.validate_token(token)
            if decoded_token is None:
                return None

            user_email = decoded_token.get("email")
            if user_email is None:
                LOGGER.warning("Email field is missing from the identity token")
                return None

            if (
                self._config.allowed_email_domain is not None
                and not user_email.endswith(self._config.allowed_email_domain)
            ):
                LOGGER.warning(
                    f"User tried to login with not allowed email address: {user_email}"
                )
                return None

            return user_email
        except Exception as e:
            LOGGER.warning(f"Failed to validate token: {str(e)}")
            return None

    def setup_authorizer(self, server: flask.Flask):
        @server.before_request
        def authorize_request():
            request = flask.request

            if self._config.oauth and request.path == P.REDIRECT_TO_LOGIN_URL:
                resp = self._redirect(self._config.oauth.sign_in_url)
                return resp

            if self._config.oauth and request.path == P.OAUTH_CODE_URL:
                code = self._get_oauth_code()
                if code is not None:
                    LOGGER.debug(f"Redirected with code={code}")
                    try:
                        id_token = self._get_identity_token(code)
                        redirect_url = flask.request.cookies.get(
                            self._config.redirect_cookie_name, MITZU_WEBAPP_URL
                        )

                        user_email = self._validate_foreign_token(id_token)
                        if not user_email:
                            raise Exception("Unauthorized (Invalid jwt token)")

                        token = self._generate_new_token_for_identity(user_email)

                        resp = self._redirect(redirect_url)
                        resp.set_cookie(self._config.token_cookie_name, token)
                        resp.set_cookie(
                            self._config.redirect_cookie_name, "", expires=0
                        )
                        return resp
                    except Exception as exc:
                        traceback.print_exception(type(exc), exc, exc.__traceback__)
                        LOGGER.warning(f"Failed to authenticate: {str(exc)}")
                        return self._get_unauthenticated_response()

            auth_token = flask.request.cookies.get(self._config.token_cookie_name)

            if request.path == P.SIGN_OUT_URL:
                if self._config.oauth and self._config.oauth.sign_out_url:
                    resp = self._redirect(self._config.oauth.sign_out_url)
                    resp.set_cookie(self._config.token_cookie_name, "", expires=0)
                    return resp
                return self._get_unauthenticated_response()

            for prefix in self._authorized_url_prefixes:
                if request.path.startswith(prefix):
                    return None

            if auth_token and self._validate_token(auth_token) is not None:
                return None

            redirect_url = request.path
            if len(request.query_string) > 0:
                redirect_url += "?" + request.query_string.decode("utf-8")
            return self._get_unauthenticated_response(redirect_url)

        @server.after_request
        def after_request(resp: flask.Response):
            request = flask.request
            for prefix in self._ignore_token_refresh_prefixes:
                if request.path.startswith(prefix):
                    return resp

            auth_token = flask.request.cookies.get(self._config.token_cookie_name)
            if auth_token is not None:
                identity = self._validate_token(auth_token)
                if identity is not None:
                    new_token = self._generate_new_token_for_identity(
                        identity["sub"], role=identity["rol"]
                    )
                    resp.set_cookie(self._config.token_cookie_name, new_token)
            return resp

    def is_request_authorized(self, request: flask.Request) -> bool:
        auth_token = request.cookies.get(self._config.token_cookie_name)
        return auth_token is not None and self._validate_token(auth_token) is not None

    def get_current_user_role(self, request: flask.Request) -> Optional[U.Role]:
        auth_token = request.cookies.get(self._config.token_cookie_name)
        if self._config.user_service is None:
            return U.Role.MEMBER

        if auth_token is None:
            return None

        claims = self._validate_token(auth_token)
        if claims is None or "rol" not in claims.keys():
            return None

        return U.Role(claims["rol"])

    def login_local_user(self, email: str, password: str) -> bool:
        if self._config.user_service is None:
            raise ValueError("User service is not set for local auth")

        user = self._config.user_service.get_user_by_email_and_password(email, password)
        if user is None:
            return False

        token = self._generate_new_token_for_identity(user.id, role=user.role)
        dash.callback_context.response.set_cookie(self._config.token_cookie_name, token)
        return True

    def get_current_user_id(self) -> Optional[str]:
        auth_token = flask.request.cookies.get(self._config.token_cookie_name)
        if auth_token is None:
            return None
        token_claims = self._validate_token(auth_token)
        if token_claims is None:
            return None
        return token_claims.get("sub")



from __future__ import annotations

from typing import Any, Dict, List, Optional

import dash.development.base_component as bc
import mitzu.model as M
import mitzu.webapp.pages.explore.complex_segment_handler as CS
import mitzu.webapp.pages.explore.metric_type_handler as MNB
from dash import html
from mitzu.webapp.helper import CHILDREN, METRIC_SEGMENTS


def from_metric(
    metric: Optional[M.Metric],
    discovered_project: M.DiscoveredProject,
) -> bc.Component:
    segments = []
    if isinstance(metric, M.SegmentationMetric):
        limit = 1
        segments = [metric._segment]
        metric_type = MNB.MetricType.SEGMENTATION
    elif isinstance(metric, M.ConversionMetric):
        limit = 10
        segments = metric._conversion._segments
        metric_type = MNB.MetricType.CONVERSION
    elif isinstance(metric, M.RetentionMetric):
        limit = 2
        segments = [metric._initial_segment, metric._retaining_segment]
        metric_type = MNB.MetricType.RETENTION
    elif metric is None:
        limit = 1
        segments = []
        metric_type = MNB.MetricType.SEGMENTATION

    fixed_metric_comps = []
    for funnel_step, segment in enumerate(segments):
        fixed_metric_comps.append(
            CS.from_segment(
                funnel_step=funnel_step,
                segment=segment,
                discovered_project=discovered_project,
                metric=metric,
                metric_type=metric_type,
            )
        )

    if len(fixed_metric_comps) < limit:
        fixed_metric_comps.append(
            CS.from_segment(
                discovered_project,
                len(fixed_metric_comps),
                None,
                None,
                metric_type,
            )
        )

    return html.Div(
        id=METRIC_SEGMENTS,
        children=fixed_metric_comps,
        className=METRIC_SEGMENTS,
    )


def from_all_inputs(
    discovered_project: Optional[M.DiscoveredProject],
    all_inputs: Dict[str, Any],
) -> List[M.Segment]:
    res: List[M.Segment] = []
    if discovered_project is None:
        return res
    complex_segments = all_inputs.get(METRIC_SEGMENTS, {}).get(CHILDREN, {})
    for _, complex_segment in complex_segments.items():
        csh = CS.from_all_inputs(discovered_project, complex_segment)
        if csh is not None:
            res.append(csh)
    return res



from __future__ import annotations

from abc import ABC
from typing import Any, Dict, List

import mitzu.model as M
import pandas as pd

# Final Select Columns
EVENT_NAME_ALIAS_COL = "_event_name"
DATETIME_COL = "_datetime"
RETENTION_INDEX = "_ret_index"
GROUP_COL = "_group"
AGG_VALUE_COL = "_agg_value"
USER_COUNT_COL = "_user_count"


# CTE Colmns
CTE_USER_ID_ALIAS_COL = "_cte_user_id"
CTE_DATETIME_COL = "_cte_datetime"
CTE_GROUP_COL = "_cte_group"


class CloseConnectionException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class GenericDatasetAdapter(ABC):
    def __init__(self, project: M.Project):
        self.project = project

    def execute_query(self, query: Any) -> pd.DataFrame:
        raise NotImplementedError()

    def list_fields(self, event_data_table: M.EventDataTable) -> List[M.Field]:
        """Returns all fields including structs and map keys for an Event Data Table
        It requires running SQL query for Map type discovery
        """
        raise NotImplementedError()

    def list_all_table_columns(self, schema: str, table_name: str) -> List[M.Field]:
        """Returns physical columns including structs for a table"""
        raise NotImplementedError()

    def get_distinct_event_names(self, event_data_table: M.EventDataTable) -> List[str]:
        raise NotImplementedError()

    def get_field_enums(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> Dict[str, M.EventDef]:
        raise NotImplementedError()

    def list_schemas(self) -> List[str]:
        raise NotImplementedError()

    def list_tables(self, schema: str) -> List[str]:
        raise NotImplementedError()

    def get_conversion_sql(self, metric: M.ConversionMetric) -> str:
        raise NotImplementedError()

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        raise NotImplementedError()

    def get_segmentation_sql(self, metric: M.SegmentationMetric) -> str:
        raise NotImplementedError()

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        raise NotImplementedError()

    def get_retention_sql(self, metric: M.RetentionMetric) -> str:
        raise NotImplementedError()

    def get_retention_df(self, metric: M.RetentionMetric) -> pd.DataFrame:
        raise NotImplementedError()

    def test_connection(self):
        raise NotImplementedError()

    def stop_current_execution(self):
        raise NotImplementedError()



from __future__ import annotations

import base64
import json
import pickle
import mitzu.model as M
from typing import Dict, Any


DISCOVERED_PROJECT_FILE_VERSION = 3

DEFS = "defs"
PROJECT = "project"
CONNECTION = "connection"


class DiscoveredProjectSerializationError(Exception):
    pass


def serialize_discovered_project(dp: M.DiscoveredProject) -> str:
    definitions = {}
    for edt, defs in dp.definitions.items():
        edt_def = {}
        for evt_name, evt_def in defs.items():
            edt_def[evt_name] = evt_def.get_value_if_exists()
        definitions[edt] = edt_def

    serializable = {
        DEFS: definitions,
        PROJECT: dp.project,
        CONNECTION: dp.project.connection,
    }

    project_binary = pickle.dumps(serializable)
    data = {
        "version": DISCOVERED_PROJECT_FILE_VERSION,
        "project": base64.b64encode(project_binary).decode("UTF-8"),
    }
    return json.dumps(data)


def deserialize_discovered_project(raw_data: bytes) -> M.DiscoveredProject:
    try:
        data = json.loads(raw_data)
        if data["version"] != DISCOVERED_PROJECT_FILE_VERSION:
            raise DiscoveredProjectSerializationError(
                "Invalid discovered project version. Please discover the project again."
            )
    except Exception as e:
        raise DiscoveredProjectSerializationError(
            "Something went wrong, cannot deserialize discovered project file.\n"
            "Try discovering the project again."
        ) from e

    res: Dict[str, Any] = pickle.loads(base64.b64decode(data["project"]))

    project: M.Project = res[PROJECT]

    for edt in project.event_data_tables:
        edt.set_project(project)
    project.set_connection(res[CONNECTION])

    definitions = {}
    for edt, defs in res[DEFS].items():
        edt_def: Dict[str, M.Reference[M.EventDef]] = {}
        for evt_name, evt_def in defs.items():
            edt_def[evt_name] = M.Reference(evt_def)
        definitions[edt] = edt_def

    return M.DiscoveredProject(definitions=definitions, project=project)



from typing import List, cast
import flask
from dash import (
    register_page,
    dcc,
    callback,
    Input,
    Output,
    State,
    html,
    no_update,
    ALL,
)
import dash_bootstrap_components as dbc
import mitzu.webapp.configs as configs
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.paths as P
from mitzu.webapp.helper import create_form_property_input


INDEX_TYPE = "local_login"
INPUT_EMAIL = "email"
INPUT_PASSWORD = "password"
BUTTON_LOGIN = "login"
LOGIN_ERROR = "login_error"
LOCATION = "login_location"


def layout(**query_params):
    depenednecies: DEPS.Dependencies = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    )
    return dbc.Container(
        children=[
            dbc.Row(
                [
                    html.Img(
                        src=configs.DASH_LOGO_PATH,
                        height="100px",
                        className="logo",
                    )
                ],
                justify="center",
            ),
            html.Hr(),
            dbc.Row(
                dbc.Col(
                    login_with_local_users()
                    if depenednecies.authorizer._config.user_service
                    else login_with_oauth(),
                    class_name="mt-5",
                )
            ),
        ]
    )


def login_with_oauth() -> List:
    return [
        dbc.Row(
            [
                dbc.Button(
                    [
                        "Login",
                    ],
                    color="secondary",
                    class_name="mb-3 w-25",
                    href=P.REDIRECT_TO_LOGIN_URL,
                    external_link=True,
                ),
            ],
            justify="center",
        ),
    ]


def login_with_local_users() -> List:
    return [
        dcc.Location(id=LOCATION, refresh=True),
        create_form_property_input(
            index_type=INDEX_TYPE,
            property=INPUT_EMAIL,
            icon_cls="bi bi-envelope",
            type="text",
            required=True,
            label_lg=1,
            input_lg=3,
            justify="center",
            name="email",
        ),
        create_form_property_input(
            index_type=INDEX_TYPE,
            property=INPUT_PASSWORD,
            icon_cls="bi bi-key",
            type="password",
            required=True,
            label_lg=1,
            input_lg=3,
            justify="center",
            name="password",
        ),
        dbc.Row(
            [
                dbc.Button(
                    [
                        "Login",
                    ],
                    color="secondary",
                    class_name="mb-3 w-25",
                    external_link=True,
                    type="submit",
                    name="login",
                    id=BUTTON_LOGIN,
                ),
            ],
            justify="center",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [],
                    lg=3,
                    sm=3,
                    id=LOGIN_ERROR,
                ),
            ],
            justify="center",
        ),
    ]


@callback(
    [Output(LOCATION, "pathname"), Output(LOGIN_ERROR, "children")],
    Input(BUTTON_LOGIN, "n_clicks"),
    State({"type": INDEX_TYPE, "index": ALL}, "value"),
    prevent_initial_call=True,
)
def login(n_click: int, inputs: List[str]):
    authorizer = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).authorizer

    if authorizer is None:
        return (
            no_update,
            html.P(
                "Authorizer is not configured for local user login",
                className="color-danger lead",
            ),
        )

    if authorizer.login_local_user(inputs[0], inputs[1]):
        return (P.HOME_PATH, "")
    return (
        no_update,
        html.P(
            "Bad credentials",
            className="text-danger lead text-center",
        ),
    )


register_page(__name__, path=P.UNAUTHORIZED_URL, title="Mitzu", layout=layout)



import dash_bootstrap_components as dbc
from dash import (
    dcc,
    ALL,
    Input,
    Output,
    State,
    callback,
    ctx,
    html,
    register_page,
    no_update,
)
import dash.development.base_component as bc
import mitzu.webapp.navbar as NB
import mitzu.webapp.pages.paths as P
import mitzu.webapp.storage as S
import mitzu.webapp.model as WM
from typing import List
import mitzu.webapp.dependencies as DEPS
from mitzu.webapp.auth.decorator import restricted, restricted_layout
import flask
from typing import cast
import datetime
from mitzu.webapp.helper import MISSING_RESOURCE_CSS
from urllib.parse import quote

CREATE_PROJECT_DOCS_LINK = "https://github.com/mitzu-io/mitzu/blob/main/DOCS.md"
SAVE_BUTTON = "project_save_button"
CLOSE_BUTTON = "project_close_button"
MANAGE_PROJECT_INFO = "manage_project_info"
SAVED_METRIC_DELETE = "save_metric_closed"
SAVED_METRICS_CONTAINER = "saved_metrics_container"
SAVED_METRICS_INFO = "saved_metrics_info"

CONFIRM_DIALOG_INDEX = "saved_metric_confirm"
CONFIRM_DIALOG_CLOSE = "saved_metric_confirm_dialog_close"
CONFIRM_DIALOG_ACCEPT = "saved_metric_confirm_dialog_accept"
CONFIRM_MODEL_BODY = "confirm_modal_body"
CONFIRM_METRIC_ID = "confirm_metric_id"

SEPARATOR = "###"


def create_confirm_dialog():
    return dbc.Modal(
        [
            dbc.ModalBody(
                "Do you really want to delete this saved metric?",
                class_name="lead",
                id=CONFIRM_MODEL_BODY,
            ),
            dbc.ModalFooter(
                [
                    dbc.Button(
                        "Close",
                        id=CONFIRM_DIALOG_CLOSE,
                        size="sm",
                        color="secondary",
                        class_name="me-1",
                    ),
                    dbc.Button(
                        "Delete",
                        id=CONFIRM_DIALOG_ACCEPT,
                        size="sm",
                        color="danger",
                    ),
                ]
            ),
            html.Div(children="", className="d-none", id=CONFIRM_METRIC_ID),
        ],
        id=CONFIRM_DIALOG_INDEX,
        is_open=False,
    )


def create_saved_metric_card(
    saved_metric: WM.SavedMetric,
) -> bc.Component:
    metric = saved_metric.metric
    project = saved_metric.project

    if project is not None and metric is not None:
        # Copying metric without ID and name for further exploration
        none_saved_params = f"?m={quote(saved_metric.metric_json)}"
        none_saved_href = f"{P.create_path(P.PROJECTS_EXPLORE_PATH, project_id=project.id)}{none_saved_params}"

        # Creating href for edit button
        saved_params = f"?sm={saved_metric.id}"
        saved_href = f"{P.create_path(P.PROJECTS_EXPLORE_PATH, project_id=project.id)}{saved_params}"
    else:
        saved_href = ""
        none_saved_href = ""

    return dbc.Card(
        [
            dbc.Row(
                [
                    dbc.Col(
                        dcc.Link(
                            children=[
                                dbc.CardImg(
                                    src=saved_metric.small_base64,
                                    className="img-fluid rounded-start",
                                    style={"width": "120px", "height": "100px"},
                                ),
                            ],
                            href=none_saved_href,
                        ),
                        className="col-md-4 ps-3",
                    ),
                    dbc.Col(
                        dbc.CardBody(
                            [
                                html.Div(
                                    [
                                        dcc.Link(
                                            html.H5(
                                                saved_metric.name,
                                                className="card-title",
                                            ),
                                            href=none_saved_href,
                                            className="text-dark",
                                        ),
                                        dbc.Button(
                                            html.I(className="bi bi-x-lg"),
                                            color="dark",
                                            id={
                                                "type": SAVED_METRIC_DELETE,
                                                "index": f"{saved_metric.id}{SEPARATOR}{saved_metric.name}",
                                            },
                                            size="sm",
                                            outline=True,
                                            className="border-0 align-top",
                                        ),
                                    ],
                                    className="d-flex justify-content-between",
                                ),
                                dcc.Link(
                                    children=[
                                        html.P(
                                            saved_metric.description,
                                            className="card-text",
                                        ),
                                        html.Small(
                                            saved_metric.created_at.strftime("%c"),
                                            className="card-text text-muted",
                                        ),
                                    ],
                                    href=none_saved_href,
                                    className="text-dark",
                                ),
                            ]
                        ),
                        className="col-md-8",
                    ),
                    dbc.Button(
                        [html.B(className="bi bi-gear me-1"), "Edit"],
                        size="sm",
                        color="light",
                        href=saved_href,
                        class_name="text-align-end",
                        disabled=project is None,
                    ),
                ],
                className="g-0 d-flex align-items-center",
            )
        ],
        class_name=MISSING_RESOURCE_CSS if project is None else "",
    )


def list_saved_metrics(storage: S.MitzuStorage) -> List[bc.Component]:
    saved_metric_ids = storage.list_saved_metrics()
    saved_metrics = [storage.get_saved_metric(s_id) for s_id in saved_metric_ids]
    saved_metrics.sort(key=lambda m: -datetime.datetime.timestamp(m.created_at))
    if len(saved_metrics) == 0:
        return dbc.Col(
            [
                html.H4("You don't have any saved metrics yet."),
                dbc.Button(
                    "Explore projects",
                    href=P.PROJECTS_PATH,
                    size="sm",
                    class_name="mt-3",
                ),
            ]
        )

    return [
        dbc.Col(
            create_saved_metric_card(sm),
            class_name="mb-3",
            sm=12,
            md=6,
            xl=4,
            lg=5,
            xs=12,
        )
        for sm in saved_metrics
    ]


@restricted_layout
def layout(**query_params) -> bc.Component:
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage

    return html.Div(
        [
            NB.create_mitzu_navbar("saved_metrics_navbar", [], storage),
            dbc.Container(
                [
                    html.H4("Saved metrics"),
                    html.Hr(),
                    html.Div("", id=SAVED_METRICS_INFO, className="lead"),
                    dbc.Row(
                        list_saved_metrics(storage),
                        id=SAVED_METRICS_CONTAINER,
                    ),
                    create_confirm_dialog(),
                ]
            ),
        ]
    )


register_page(
    __name__,
    path=P.SAVED_METRICS,
    title="Mitzu - Saved Metrics",
    layout=layout,
)


@callback(
    Output(CONFIRM_DIALOG_INDEX, "is_open"),
    Output(CONFIRM_METRIC_ID, "children"),
    Output(CONFIRM_MODEL_BODY, "children"),
    Input({"type": SAVED_METRIC_DELETE, "index": ALL}, "n_clicks"),
    Input(CONFIRM_DIALOG_CLOSE, "n_clicks"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    prevent_initial_call=True,
)
@restricted
def delete_button_clicked(delete_clicks: List[int], close: int, accept: int):
    if ctx.triggered_id is None or all([v is None for v in delete_clicks]):
        return no_update, no_update, no_update

    if ctx.triggered_id in [CONFIRM_DIALOG_CLOSE, CONFIRM_DIALOG_ACCEPT]:
        return (False, "", "")

    index = ctx.triggered_id.get("index", "")
    parts = index.split(SEPARATOR)

    if len(parts) == 2:
        metric_id, metric_name = tuple(parts)
        message = f"Do you really want to delete {metric_name}"
        return (True, metric_id, message)

    return (False, "", "")


@callback(
    Output(SAVED_METRICS_CONTAINER, "children"),
    Input(CONFIRM_DIALOG_ACCEPT, "n_clicks"),
    State(CONFIRM_METRIC_ID, "children"),
    background=True,
    running=[
        (
            Output(SAVED_METRICS_INFO, "children"),
            [
                dbc.Spinner(
                    spinner_style={"width": "1rem", "height": "1rem"},
                    spinner_class_name="me-1",
                ),
                "Loading saved metrics",
            ],
            "",
        )
    ],
    interval=200,
    prevent_initial_call=True,
)
@restricted
def confirm_button_clicked(close_button: int, metric_id: str) -> List:
    storage = cast(
        DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
    ).storage
    if ctx.triggered_id is not None:
        storage.clear_saved_metric(metric_id)

    return list_saved_metrics(storage)



from __future__ import annotations

import os
from typing import List, Optional
from mitzu.webapp.auth.authorizer import (
    OAuthConfig,
)


class Cognito:
    @classmethod
    def get_config(
        cls,
        pool_id: Optional[str] = None,
        region: Optional[str] = None,
        domain: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_url: Optional[str] = None,
        jwt_algo: List[str] = ["RS256"],
    ) -> OAuthConfig:
        """
        Creates a new OAuthConfig instance using the Cognito specific settings.

        :param pool_id: ID of the AWS Cognito User Pool (something like ``<aws region>_xxxxxxxxx``),
            if it's not set then the ``COGNITO_POOL_ID`` environmental variable will be used
        :param region: AWS region of the AWS Cognito User Pool (eg. eu-west-1),
            if it's not set then the ``COGNITO_REGION`` environmental variable will be used
        :param domain: AWS Cognito domain set for the App integrations
            (``xxxxx`` from the ``https://xxxxx.auth.eu-west-1.amazoncognito.com``),
            if it's not set the the ``COGNITO_DOMAIN`` environmental variable will be used
        :param client_id: Client ID of the configured App,
            if it's not set then the ``COGNITO_CLIENT_ID`` environmental variable will be used
        :param client_secret: Client secret of the configured App,
            if it's not set then the ``COGNITO_CLIENT_SECRET`` environmental variable will be used
        :param redirect_url: OAuth redirection URL, it's ``<your Mitzu base URL>/auth/oauth`` eg. ``http://localhost:8082/auth/oauth``,
            if it's not set then the ``COGNITO_REDIRECT_URL`` environmental variable will be used
        :param jwt_algo: List of accepted JWT token signing algorithms, by default it's ``["RS256"]``
            it can be overwritted with the ``COGNITO_JWT_ALGORITHMS`` environmental variable
        """
        pool_id = cls.fallback_to_env_var(pool_id, "COGNITO_POOL_ID")
        region = cls.fallback_to_env_var(region, "COGNITO_REGION")
        domain = cls.fallback_to_env_var(domain, "COGNITO_DOMAIN")
        client_id = cls.fallback_to_env_var(client_id, "COGNITO_CLIENT_ID")
        redirect_url = cls.fallback_to_env_var(redirect_url, "COGNITO_REDIRECT_URL")
        jwt_algorithms = cls.fallback_to_env_var(
            ",".join(jwt_algo), "COGNITO_JWT_ALGORITHMS"
        ).split(",")

        return OAuthConfig(
            client_id=client_id,
            client_secret=cls.fallback_to_env_var(
                client_secret, "COGNITO_CLIENT_SECRET"
            ),
            jwks_url=f"https://cognito-idp.{region}.amazonaws.com/{pool_id}/.well-known/jwks.json",
            sign_in_url=(
                f"https://{domain}/oauth2/authorize?"
                f"client_id={client_id}&"
                "response_type=code&"
                "scope=email+openid&"
                f"redirect_uri={redirect_url}"
            ),
            sign_out_url=(
                f"https://{domain}/logout?"
                f"client_id={client_id}&"
                "response_type=code&"
                "scope=email+openid&"
                f"redirect_uri={redirect_url}"
            ),
            token_url=f"https://{domain}/oauth2/token",
            jwt_algorithms=jwt_algorithms,
        )

    @classmethod
    def fallback_to_env_var(cls, value: Optional[str], env_var: str):
        if value is not None:
            return value
        return os.getenv(env_var)



import math
import random
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple
from uuid import uuid4

import pandas as pd

START_DATE = datetime(2021, 1, 1)
LENGTH = 365

texts = [
    "jeans",
    "shirt",
    "blue",
    "red",
    "yellow",
    "black",
    "gray",
    "pink",
    "shoes",
    "jumper",
    "long",
    "short",
]

country_codes = [
    "cn",
    "us",
    "de",
    "fr",
    "hu",
    "gb",
    "br",
]

locales = [
    "cn_CN",
    "en_US",
    "de_DE",
    "fr_FR",
    "hu_HU",
    "en_GB",
    "br_BR",
]

campaigns = [
    "summer_sale",
    "promo_20off",
    "christmass_sale",
    "organic",
    "facebook_amer",
    "google_ads_global",
]


subscription_reason = ["promo_code", "organic"]

page_domains = [
    "awesomestore.com",
    "awesomestore.uk",
    "megamagasin.com",
    "awesomestore.fr",
]

page_titles = ["home", "search", "blogs_page", "deals_page"]


def random_user() -> Dict:
    return {
        "user_id": str(uuid4()),
        "user_country_code": random.choice(country_codes),
        "user_locale": random.choice(locales),
    }


def random_datetime(freq: float, phase: float) -> datetime:
    while True:
        days_to_add = random.randrange(0, LENGTH)
        if (
            random.random() < math.sin(freq * (days_to_add / 180) + phase) + 1
            and round(abs(days_to_add % 7 - 3.5)) < random.random() * 14
        ):
            break

    return START_DATE + timedelta(days=days_to_add, seconds=random.randrange(0, 86400))


def random_timedelta(range_secs: int, range_days: int = 0) -> timedelta:
    return timedelta(
        seconds=random.random() ** 2 * range_secs,
        days=random.random() ** 2 * range_days,
    )


def page_event(
    event_time: datetime,
    user: Dict,
    event_name: str,
    domain: str,
    title: str,
    item_id: Optional[str],
    acq_campaign: str,
) -> Dict:
    return {
        "event_name": event_name,
        "event_time": event_time,
        "title": title,
        "domain": domain,
        "acquisition_campaign": acq_campaign,
        "item_id": item_id,
        **user,
    }


def search_event(
    event_time: datetime,
    user: Dict,
    search_text: str,
) -> Dict:
    return {
        "event_name": "search",
        "event_time": event_time,
        "search_text": search_text,
        **user,
    }


def add_to_cart(
    event_time: datetime,
    user: Dict,
    item_id: str,
) -> Dict:
    return {
        "event_name": "add_to_cart",
        "event_time": event_time,
        "item": item_id,
        **user,
    }


def checkout(event_time: datetime, user: Dict, cost: int, num_items: int) -> Dict:
    return {
        "event_name": "checkout",
        "event_time": event_time,
        "cost_usd": cost,
        "number_of_items": num_items,
        **user,
    }


def subscribe(event_time: datetime, user: Dict, email_campaign: str) -> Dict:
    return {
        "event_name": "subscribe",
        "event_time": event_time,
        "email_campaign": email_campaign,
        **user,
    }


def email_sent(event_time: datetime, user: Dict, deal_id: str) -> Dict:
    return {
        "event_name": "email_sent",
        "event_time": event_time,
        "deal_id": deal_id,
        **user,
    }


def email_opened(event_time: datetime, user: Dict) -> Dict:
    return {
        "event_name": "email_opened",
        "event_time": event_time,
        **user,
    }


ISO_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"


@dataclass(frozen=True)
class FunnelDistribution:
    funnels_count: int
    frequency: float
    phase: float
    users: List[Dict]
    page_domain: str
    page_title: str

    acquisition_campaign: str
    churn_probability: float
    add_to_cart_probability: float
    email_open_probability: float
    email_campaign: str


def create_user_funnel(
    dist: FunnelDistribution,
    user: Dict,
    page_events: List[Dict],
    search_events: List[Dict],
    add_to_carts: List[Dict],
    checkouts: List[Dict],
    email_subscriptions: List[Dict],
    email_sent_events: List[Dict],
    email_opened_events: List[Dict],
):
    # page_visit
    event_time = random_datetime(dist.frequency, dist.phase)
    page_events.append(
        page_event(
            event_time=event_time,
            event_name="page_visit",
            user=user,
            domain=dist.page_domain,
            title=dist.page_title,
            item_id=None,
            acq_campaign=dist.acquisition_campaign,
        )
    )
    if random.random() < dist.churn_probability:
        page_events.append(
            page_event(
                event_time=event_time,
                event_name="element_clicked_on_page",
                user=user,
                domain=dist.page_domain,
                title=dist.page_title,
                item_id=str(random.randrange(1, 100)),
                acq_campaign=dist.acquisition_campaign,
            )
        )

    if random.random() > dist.churn_probability:
        return

    # search + add_to_cart
    for _ in range(0, int(random.random() ** 2 * 10)):
        event_time = event_time + random_timedelta(180)
        search_events.append(
            search_event(
                event_time=event_time,
                user=user,
                search_text=" ".join(
                    [random.choice(texts) for i in range(1, random.randint(1, 4))]
                ),
            )
        )

        if random.random() < dist.add_to_cart_probability:
            add_to_carts.append(
                add_to_cart(
                    event_time=event_time,
                    user=user,
                    item_id=f"item-{random.randrange(0, 10000)}",
                )
            )

    if random.random() < dist.churn_probability:
        return

    # checkout
    event_time = event_time + random_timedelta(300)
    checkouts.append(
        checkout(
            event_time=event_time,
            user=user,
            cost=int(random.random() ** 2 * 2000),
            num_items=int(random.random() ** 2 * 20),
        )
    )

    if random.random() < dist.churn_probability:
        return

    event_time = event_time + random_timedelta(300)
    checkouts.append(
        checkout(
            event_time=event_time,
            user=user,
            cost=int(random.random() ** 2 * 2000),
            num_items=int(random.random() ** 2 * 20),
        )
    )

    if random.random() < dist.churn_probability:
        return

    event_time = event_time + random_timedelta(300)
    email_subscriptions.append(
        subscribe(event_time=event_time, user=user, email_campaign=dist.email_campaign)
    )

    # email_sent + email_opened
    for _ in range(0, int(random.random() ** 2 * 10)):
        event_time = event_time + random_timedelta(range_days=14, range_secs=86400)
        email_sent_events.append(
            email_sent(
                event_time=event_time,
                user=user,
                deal_id=f"{dist.email_campaign}-{random.randrange(0,100)}",
            )
        )
        if random.random() < dist.email_open_probability:
            event_time = event_time + random_timedelta(range_days=3, range_secs=0)
            email_opened_events.append(email_opened(event_time=event_time, user=user))
            if random.random() < dist.churn_probability:
                break


def generate_checkout_funnels(
    dist: FunnelDistribution,
) -> Tuple[Dict[str, pd.DataFrame], int]:
    page_events: List[Dict] = []
    search_events: List[Dict] = []
    add_to_carts: List[Dict] = []
    checkouts: List[Dict] = []
    email_subscriptions: List[Dict] = []
    email_sent_events: List[Dict] = []
    email_opened_events: List[Dict] = []

    for _ in range(0, dist.funnels_count):
        user = random.choice(dist.users)
        create_user_funnel(
            dist=dist,
            user=user,
            page_events=page_events,
            search_events=search_events,
            add_to_carts=add_to_carts,
            checkouts=checkouts,
            email_opened_events=email_opened_events,
            email_sent_events=email_sent_events,
            email_subscriptions=email_subscriptions,
        )

    created_count = (
        len(page_titles)
        + len(page_events)
        + len(search_events)
        + len(add_to_carts)
        + len(checkouts)
        + len(email_subscriptions)
        + len(email_sent_events)
        + len(email_opened_events)
    )

    return (
        {
            "page_events": pd.DataFrame(page_events),
            "search_events": pd.DataFrame(search_events),
            "add_to_carts": pd.DataFrame(add_to_carts),
            "checkouts": pd.DataFrame(checkouts),
            "email_subscriptions": pd.DataFrame(email_subscriptions),
            "email_sent_events": pd.DataFrame(email_sent_events),
            "email_opened_events": pd.DataFrame(email_opened_events),
        },
        created_count,
    )


def create_all_funnels(
    user_count: int = 1000, event_count: int = 100000, seed: Optional[int] = 100
) -> Dict[str, pd.DataFrame]:
    if seed is not None:
        random.seed(seed)
    results_df: Dict[str, pd.DataFrame] = {}
    users = [random_user() for _ in range(0, user_count)]

    while event_count > 0:
        dist = FunnelDistribution(
            frequency=random.random() * 3,
            phase=random.random() * 3.6,
            users=users,
            page_domain=random.choice(page_domains),
            page_title=random.choice(page_titles),
            acquisition_campaign=random.choice(campaigns),
            churn_probability=random.random(),
            add_to_cart_probability=random.random(),
            email_open_probability=0.3,
            email_campaign=random.choice(campaigns),
            funnels_count=random.randrange(0, 10000),
        )
        dfs, created_event_cnt = generate_checkout_funnels(dist)
        event_count -= created_event_cnt
        for key, df in dfs.items():
            if key in results_df:
                df = pd.concat([results_df[key], df])
            results_df[key] = df

    return results_df



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mitzu.adapters.sqlalchemy.databricks.sqlalchemy.error import DatabricksQueryError

from sqlalchemy.sql import compiler

# https://docs.databricks.com/spark/latest/spark-sql/language-manual/sql-ref-reserved-words.html
DB_RESERVED_WORDS = {
    #  reserves words
    "anti",
    "cross",
    "except",
    "full",
    "inner",
    "intersect",
    "join",
    "lateral",
    "left",
    "minus",
    "natural",
    "on",
    "right",
    "semi",
    "union",
    "using",
    # expression special words
    "null",
    "default",
    # reserved scheman words
    "builtin",
    "session",
    "information_schema",
    "sys",
    # other
    "all",
    "alter",
    "and",
    "any",
    "array",
    "as",
    "at",
    "authorization",
    "between",
    "both",
    "by",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "commit",
    "constraint",
    "create",
    "cross",
    "cube",
    "current",
    "current_date",
    "current_time",
    "current_timestamp",
    "current_user",
    "delete",
    "describe",
    "distinct",
    "drop",
    "else",
    "end",
    "escape",
    "except",
    "exists",
    "external",
    "extract",
    "false",
    "fetch",
    "filter",
    "for",
    "foreign",
    "from",
    "full",
    "function",
    "global",
    "grant",
    "group",
    "grouping",
    "having",
    "in",
    "inner",
    "insert",
    "intersect",
    "interval",
    "into",
    "is",
    "join",
    "leading",
    "left",
    "like",
    "local",
    "natural",
    "no",
    "not",
    "null",
    "of",
    "on",
    "only",
    "or",
    "order",
    "out",
    "outer",
    "overlaps",
    "partition",
    "position",
    "primary",
    "range",
    "references",
    "revoke",
    "right",
    "rollback",
    "rollup",
    "row",
    "rows",
    "select",
    "session_user",
    "set",
    "some",
    "start",
    "table",
    "tablesample",
    "then",
    "time",
    "to",
    "trailing",
    "true",
    "truncate",
    "union",
    "unique",
    "unknown",
    "update",
    "user",
    "using",
    "values",
    "when",
    "where",
    "window",
    "with",
}


class DatabricksSQLCompiler(compiler.SQLCompiler):
    def limit_clause(self, select, **kw):
        if select._offset_clause is not None:
            raise DatabricksQueryError("OFFSET is not supported")
        if select._limit_clause is not None:
            return "\nLIMIT " + self.process(select._limit_clause, **kw)
        return ""


class DatabricksDDLCompiler(compiler.DDLCompiler):
    pass


class DatabricksTypeCompiler(compiler.GenericTypeCompiler):
    def visit_FLOAT(self, type_, **kw):
        precision = type_.precision or 32
        if 0 <= precision <= 32:
            return self.visit_REAL(type_, **kw)
        elif 32 < precision <= 64:
            return self.visit_DOUBLE(type_, **kw)
        else:
            raise ValueError(
                f"type.precision must be in range [0, 64], got {type_.precision}"
            )

    def visit_DOUBLE(self, type_, **kw):
        return "DOUBLE"

    def visit_NUMERIC(self, type_, **kw):
        return self.visit_DECIMAL(type_, **kw)

    def visit_NCHAR(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_NVARCHAR(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_TEXT(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_BINARY(self, type_, **kw):
        return self.visit_VARBINARY(type_, **kw)

    def visit_CLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_NCLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_BLOB(self, type_, **kw):
        return self.visit_VARCHAR(type_, **kw)

    def visit_DATETIME(self, type_, **kw):
        return self.visit_TIMESTAMP(type_, **kw)


class DatabricksIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words = DB_RESERVED_WORDS

    def __init__(
        self,
        dialect,
        initial_quote='"',
        final_quote=None,
        escape_quote='"',
        quote_case_sensitive_collations=True,
        omit_schema=False,
    ):
        super().__init__(
            dialect=dialect,
            initial_quote=initial_quote,
            final_quote=final_quote,
            escape_quote=escape_quote,
            quote_case_sensitive_collations=quote_case_sensitive_collations,
            omit_schema=omit_schema,
        )



from __future__ import annotations

from typing import Any, List

import mitzu.adapters.generic_adapter as GA
import mitzu.model as M
import pandas as pd
from mitzu.adapters.helper import dataframe_str_to_datetime
from mitzu.adapters.sqlalchemy_adapter import (
    FieldReference,
    SQLAlchemyAdapter,
)

import sqlalchemy as SA
import sqlalchemy.sql.expression as EXP

ARRAY_JOIN_SEP = "###"


class SQLiteAdapter(SQLAlchemyAdapter):
    def __init__(self, project: M.Project):
        super().__init__(project)

    def _column_index_support(self):
        return False

    def _get_connection_url(self, con: M.Connection):
        if not con.host:
            return "sqlite://?check_same_thread=False"
        else:
            return f"sqlite:///{con.host}.db?check_same_thread=False"

    def keep_alive_connection(self) -> bool:
        return True

    def _get_date_trunc(self, time_group: M.TimeGroup, field_ref: FieldReference):
        if time_group == M.TimeGroup.WEEK:
            return SA.func.datetime(SA.func.date(field_ref, "weekday 0", "-6 days"))
        if time_group == M.TimeGroup.SECOND:
            fmt = "%Y-%m-%dT%H:%M:%S"
        elif time_group == M.TimeGroup.MINUTE:
            fmt = "%Y-%m-%dT%H:%M:00"
        elif time_group == M.TimeGroup.HOUR:
            fmt = "%Y-%m-%dT%H:00:00"
        elif time_group == M.TimeGroup.DAY:
            fmt = "%Y-%m-%dT00:00:00"
        elif time_group == M.TimeGroup.MONTH:
            fmt = "%Y-%m-01T00:00:00"
        elif time_group == M.TimeGroup.QUARTER:
            raise NotImplementedError(
                "Timegroup Quarter is not supported for SQLite Adapter"
            )
        elif time_group == M.TimeGroup.YEAR:
            fmt = "%Y-01-01T00:00:00"

        return SA.func.datetime(SA.func.strftime(fmt, field_ref))

    def _get_column_values_df(
        self,
        event_data_table: M.EventDataTable,
        fields: List[M.Field],
        event_specific: bool,
    ) -> pd.DataFrame:
        df = super()._get_column_values_df(event_data_table, fields, event_specific)
        for field in df.columns:
            df[field] = (
                df[field]
                .str.replace(f"{ARRAY_JOIN_SEP}$", "", regex=True)
                .str.split(f"{ARRAY_JOIN_SEP},")
            )
        return df

    def _get_distinct_array_agg_func(self, field_ref: FieldReference) -> Any:
        return SA.func.group_concat(SA.distinct(field_ref.concat(ARRAY_JOIN_SEP)))

    def _get_datetime_interval(
        self, field_ref: FieldReference, timewindow: M.TimeWindow
    ) -> Any:
        if timewindow.period == M.TimeGroup.WEEK:
            # SQL Lite doesn't have the week concept
            timewindow = M.TimeWindow(timewindow.value * 7, M.TimeGroup.DAY)
        return SA.func.datetime(
            field_ref,
            SA.text(f"'+{timewindow.value} {timewindow.period.name.lower()}'"),
        )

    def _get_dynamic_datetime_interval(
        self,
        field_ref: FieldReference,
        value_field_ref: FieldReference,
        time_group: M.TimeGroup,
    ) -> Any:
        if time_group == M.TimeGroup.WEEK:
            # SQL Lite doesn't have the week concept
            time_group = M.TimeGroup.DAY
            return SA.func.datetime(
                field_ref,
                SA.text(
                    f"'+' || ({value_field_ref} * 7) || ' {time_group.name.lower()}'"
                ),
            )

        return SA.func.datetime(
            field_ref,
            SA.text(f"'+' || {value_field_ref} || ' {time_group.name.lower()}'"),
        )

    def _get_conv_aggregation(
        self, metric: M.Metric, cte: EXP.CTE, first_cte: EXP.CTE
    ) -> Any:
        if metric._agg_type == M.AggType.PERCENTILE_TIME_TO_CONV:
            raise NotImplementedError(
                "Percentile calculation is not supported by SQLite Connections."
            )
        if metric._agg_type == M.AggType.AVERAGE_TIME_TO_CONV:
            t1 = first_cte.columns.get(GA.CTE_DATETIME_COL)
            t2 = cte.columns.get(GA.CTE_DATETIME_COL)
            diff = SA.func.round(
                (SA.func.julianday(t2) - SA.func.julianday(t1)) * 86400
            )
            return SA.func.avg(diff)
        else:
            return super()._get_conv_aggregation(metric, cte, first_cte)

    def get_conversion_df(self, metric: M.ConversionMetric) -> pd.DataFrame:
        df = super().get_conversion_df(metric)
        return dataframe_str_to_datetime(df, GA.DATETIME_COL)

    def get_segmentation_df(self, metric: M.SegmentationMetric) -> pd.DataFrame:
        df = super().get_segmentation_df(metric)
        return dataframe_str_to_datetime(df, GA.DATETIME_COL)

    def get_retention_df(self, metric: M.RetentionMetric) -> pd.DataFrame:
        df = super().get_retention_df(metric)
        df = dataframe_str_to_datetime(df, GA.DATETIME_COL)
        return df

    def _get_datetime_column(self, cte: Any, name: str) -> Any:
        return SA.func.datetime(cte.columns.get(name))



from typing import cast

import dash.development.base_component as bc
import flask
from dash import html, register_page

import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.explore.explore_page as EXP
import mitzu.webapp.pages.paths as P
from mitzu.webapp.auth.decorator import restricted_layout
import traceback

register_page(
    __name__,
    path_template=P.PROJECTS_EXPLORE_PATH,
    title="Mitzu - Explore",
)


@restricted_layout
def layout(project_id: str, **query_params) -> bc.Component:
    try:
        depenednecies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )
        project = depenednecies.storage.get_project(project_id)

        if project is None:
            return html.Div("Project not found", className="d-flex text-center lead")

        discovered_project = project._discovered_project.get_value()
        if discovered_project is None:
            return html.Div(
                "Project have not been discovered yet",
                className="d-flex text-center lead",
            )

        return EXP.create_explore_page(
            query_params=query_params,
            discovered_project=discovered_project,
            storage=depenednecies.storage,
        )
    except Exception as exc:
        traceback.print_exc()
        return html.Div(str(exc))


EXP.create_callbacks()



import flask
from typing import cast
import functools
from dash.exceptions import PreventUpdate
from dash.dcc import Location
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.pages.paths as P
import mitzu.webapp.service.user_service as US


def restricted(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        dependencies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )

        if dependencies.authorizer is not None:
            if dependencies.authorizer.is_request_authorized(flask.request):
                return func(*args, **kwargs)
            else:
                raise PreventUpdate
        else:
            return func(*args, **kwargs)

    return wrapper


def restricted_for_admin(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        dependencies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )

        if dependencies.authorizer is not None:
            user_role = dependencies.authorizer.get_current_user_role(flask.request)
            if user_role is None:
                raise PreventUpdate

            if user_role != US.Role.ADMIN:
                raise PreventUpdate
            else:
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return wrapper


def restricted_layout(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        dependencies: DEPS.Dependencies = cast(
            DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
        )

        if dependencies.authorizer is not None:
            if dependencies.authorizer.is_request_authorized(flask.request):
                return func(*args, **kwargs)
            else:
                return Location("url", href=P.UNAUTHORIZED_URL)
        else:
            return func(*args, **kwargs)

    return wrapper



from mitzu.model import (
    Connection,
    ConnectionType,
    DiscoveredProject,
    EventDataTable,
    Project,
    PromptSecretResolver,
    ConstSecretResolver,
    EnvVarSecretResolver,
    TimeWindow,
    TimeGroup,
    DiscoverySettings,
    WebappSettings,
)
from mitzu.samples import get_sample_discovered_project

__version__ = "0.5.6-rc.1"


__all__ = [
    "Connection",
    "ConnectionType",
    "Project",
    "EventDataTable",
    "DiscoveredProject",
    "PromptSecretResolver",
    "ConstSecretResolver",
    "EnvVarSecretResolver",
    "TimeWindow",
    "TimeGroup",
    "get_sample_discovered_project",
    "DiscoverySettings",
    "WebappSettings",
]


def load_from_project_file(
    project: str, folder: str = "./", extension="mitzu"
) -> DiscoveredProject:
    return DiscoveredProject.load_from_project_file(project, folder, extension)



from __future__ import annotations

from typing import List, Optional

import dash.development.base_component as bc
import dash_bootstrap_components as dbc
from dash import dcc, html
import dash_mantine_components as dmc

import mitzu.model as M
from mitzu.helper import value_to_label

CHART_TYPE_DD = "chart_type_button"
CHART_TYPE_CONTAINER = "chart_type_container"
GRAPH_CONTENT_TYPE = "graph_content_type"
GRAPH_REFRESH_BUTTON = "graph_refresh_button"
GRAPH_RUN_QUERY_BUTTON = "graph_run_query_button"
CANCEL_BUTTON = "cancel_button"

TOOLBAR_ROW = "toolbar_row"

VISIBLE = {"display": "inline-block", "marginLeft": "8px"}
HIDDEN = {"display": "none"}

CHART_VAL = "Chart"
TABLE_VAL = "Table"
SQL_VAL = "SQL"


CONTENT_CLIPBOARD = "content_clipboard"


def create_chart_type_dropdown_options(
    metric: Optional[M.Metric],
) -> List[M.SimpleChartType]:
    if metric is None:
        return [M.SimpleChartType.LINE]
    if metric._time_group == M.TimeGroup.TOTAL:
        options = [
            M.SimpleChartType.BAR,
        ]
        if isinstance(metric, M.SegmentationMetric):
            options.extend([M.SimpleChartType.STACKED_BAR])
        if isinstance(metric, M.RetentionMetric):
            options.insert(0, M.SimpleChartType.LINE)
    else:
        options = [
            M.SimpleChartType.LINE,
            M.SimpleChartType.BAR,
            M.SimpleChartType.HEATMAP,
        ]
        if isinstance(metric, M.SegmentationMetric):
            options.extend(
                [
                    M.SimpleChartType.STACKED_BAR,
                    M.SimpleChartType.STACKED_AREA,
                ]
            )
    return options


def create_chart_type_dropdown(metric: Optional[M.Metric]) -> dcc.Dropdown:
    options = create_chart_type_dropdown_options(metric)
    if metric is not None and metric._chart_type in options:
        ct = metric._chart_type
    else:
        ct = options[0]

    return dmc.Select(
        data=[{"label": value_to_label(o.name), "value": o.name} for o in options],
        value=ct.name,
        id=CHART_TYPE_DD,
        size="xs",
        className="me-1 btn-secondary",
        style={
            "maxWidth": "120px",
            "display": "inline-block",
        },
    )


def from_metric(metric: Optional[M.Metric], project: M.Project) -> bc.Component:
    auto_refresh = project.webapp_settings.auto_refresh_enabled
    comp = dbc.Row(
        id=TOOLBAR_ROW,
        children=[
            html.Hr(),
            dbc.Col(
                children=[
                    dbc.Button(
                        children=[html.B(className="bi bi-arrow-clockwise")],
                        size="sm",
                        color="primary",
                        id=GRAPH_REFRESH_BUTTON,
                        disabled=False,
                        class_name="me-1 " + ("" if auto_refresh else "d-none"),
                    ),
                    dbc.Button(
                        children=[html.B(className="bi bi-play-fill me-1"), "Run"],
                        size="sm",
                        color="primary",
                        id=GRAPH_RUN_QUERY_BUTTON,
                        disabled=False,
                        class_name="me-1 " + ("d-none" if auto_refresh else ""),
                    ),
                    dbc.Button(
                        children=[
                            dbc.Spinner(
                                size="sm",
                                color="dark",
                                type="border",
                                spinnerClassName="me-1",
                            ),
                            "Cancel",
                        ],
                        size="sm",
                        color="secondary",
                        id=CANCEL_BUTTON,
                        style=HIDDEN,
                        class_name=("" if auto_refresh else "d-none"),
                    ),
                ],
            ),
            dbc.Col(
                children=[
                    html.Div(
                        children=[create_chart_type_dropdown(metric)],
                        id=CHART_TYPE_CONTAINER,
                        className="d-inline-block",
                    ),
                    dmc.SegmentedControl(
                        data=[CHART_VAL, TABLE_VAL, SQL_VAL],
                        size="xs",
                        id=GRAPH_CONTENT_TYPE,
                        className="me-1 bg-transparent",
                        value=CHART_VAL,
                    ),
                ],
            ),
        ],
        class_name="mt-3",
    )

    return comp



from __future__ import annotations


import plotly.express as px
import plotly.figure_factory as ff

import pandas as pd
import mitzu.model as M
import mitzu.visualization.common as C
from typing import Dict, List, Tuple
from base64 import b64encode
import plotly.io as pio
import traceback

PRISM2 = [
    "rgb(95, 70, 144)",
    "rgb(237, 173, 8)",
    "rgb(29, 105, 150)",
    "rgb(225, 124, 5)",
    "rgb(56, 166, 165)",
    "rgb(204, 80, 62)",
    "rgb(15, 133, 84)",
    "rgb(148, 52, 110)",
    "rgb(115, 175, 72)",
    "rgb(111, 64, 112)",
    "rgb(102, 102, 102)",
]


def set_figure_style(fig, simple_chart: C.SimpleChart, metric: M.Metric):
    if simple_chart.title is not None:
        title_height = len(simple_chart.title.split("<br />")) * 30
    else:
        title_height = 0

    if metric._chart_type == M.SimpleChartType.LINE:
        fig.update_traces(
            line=dict(width=1.5),
            mode="lines+markers",
            textposition="top center",
            textfont_size=9,
        )
    elif metric._chart_type == M.SimpleChartType.STACKED_AREA:
        fig.update_traces(textposition="top center", textfont_size=9)
    else:
        fig.update_traces(textposition="outside", textfont_size=9)

    fig.update_yaxes(
        rangemode="tozero",
        showline=True,
        linecolor="rgba(127,127,127,0.1)",
        gridcolor="rgba(127,127,127,0.1)",
        fixedrange=True,
    )
    fig.update_xaxes(
        rangemode="tozero",
        showline=True,
        linecolor="rgba(127,127,127,0.3)",
        gridcolor="rgba(127,127,127,0.3)",
        fixedrange=True,
        showgrid=False,
    )
    fig.update_layout(
        title={
            "text": simple_chart.title,
            "x": 0.5,
            "xanchor": "center",
            "yanchor": "top",
            "font": {"size": 14},
        },
        autosize=True,
        bargap=0.30,
        bargroupgap=0.15,
        margin=dict(t=title_height, l=1, r=1, b=1, pad=0),
        uniformtext_minsize=7,
        height=600,
        hoverlabel={"font": {"size": 12}},
        paper_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        hovermode=simple_chart.hover_mode,
        legend=dict(orientation="h", yanchor="bottom", y=-0.2, xanchor="left"),
        showlegend=len(fig.data) > 1,
    )
    return fig


def set_heatmap_figure_style(fig, simple_chart: C.SimpleChart):
    fig.update_traces(xgap=1, ygap=1, hovertemplate="%{hovertext} <extra></extra>")
    fig.update_xaxes(
        side="top",
        title=simple_chart.title,
        showgrid=False,
    )
    fig.update_yaxes(
        side="left",
        title=simple_chart.y_axis_label,
        showgrid=False,
    )
    fig["layout"]["yaxis"]["autorange"] = "reversed"

    if simple_chart.title is not None:
        title_height = len(simple_chart.title.split("<br />")) * 30
    else:
        title_height = 0

    fig.update_layout(
        showlegend=True,
        uniformtext_minsize=9,
        uniformtext_mode="hide",
        autosize=True,
        hoverlabel={"font": {"size": 12}},
        annotations={"font": {"size": 8}},
        hovermode="closest",
        margin=dict(t=title_height, l=1, r=1, b=1, pad=0),
    )
    return fig


def pdf_to_heatmap(pdf: pd.DataFrame) -> Dict[str, List]:
    pdf = pd.pivot_table(
        pdf,
        values=[C.TEXT_COL, C.Y_AXIS_COL, C.TOOLTIP_COL],
        index=[C.X_AXIS_COL],
        columns=[C.COLOR_COL],
        aggfunc="first",
        sort=False,
    )
    color_vals = [v for v in pdf.index.tolist()]
    x_vals = [*dict.fromkeys([f"{val[1]}." for val in pdf.columns.tolist()])]

    return {
        "x": x_vals,
        "y": color_vals,
        "z": pdf[C.Y_AXIS_COL].values.tolist(),
        "hovertext": pdf[C.TOOLTIP_COL].values.tolist(),
        "annotation_text": pdf[C.TEXT_COL].values.tolist(),
    }


def plot_chart(
    simple_chart: C.SimpleChart,
    metric: M.Metric,
):
    size = simple_chart.dataframe.shape[0]
    if size > 3000:
        raise Exception(
            f"Too many data points to visualize ({size}), try reducing the scope."
        )
    pdf = simple_chart.dataframe.copy()
    if simple_chart.x_axis_labels_func is not None:
        pdf[C.X_AXIS_COL] = pdf[C.X_AXIS_COL].apply(
            lambda val: simple_chart.x_axis_labels_func(val, metric)
        )
    if simple_chart.y_axis_labels_func is not None:
        pdf[C.Y_AXIS_COL] = pdf[C.Y_AXIS_COL].apply(
            lambda val: simple_chart.y_axis_labels_func(val, metric)
        )
    if simple_chart.color_labels_func is not None:
        pdf[C.COLOR_COL] = pdf[C.COLOR_COL].apply(
            lambda val: simple_chart.color_labels_func(val, metric)
        )

    ct = simple_chart.chart_type
    px.defaults.color_discrete_sequence = PRISM2
    if ct in [
        M.SimpleChartType.BAR,
        M.SimpleChartType.STACKED_BAR,
    ]:
        barmode = "group" if ct in [M.SimpleChartType.BAR] else "relative"
        orientation = (
            "v" if ct in [M.SimpleChartType.STACKED_BAR, M.SimpleChartType.BAR] else "h"
        )
        fig = px.bar(
            pdf,
            x=C.X_AXIS_COL,
            y=C.Y_AXIS_COL,
            text=C.TEXT_COL,
            color=C.COLOR_COL,
            barmode=barmode,
            orientation=orientation,
            custom_data=[C.TOOLTIP_COL],
            labels={
                C.X_AXIS_COL: simple_chart.x_axis_label,
                C.Y_AXIS_COL: simple_chart.y_axis_label,
                C.COLOR_COL: simple_chart.color_label,
            },
        )
    elif ct == M.SimpleChartType.LINE:
        fig = px.line(
            pdf,
            x=C.X_AXIS_COL,
            y=C.Y_AXIS_COL,
            text=C.TEXT_COL,
            color=C.COLOR_COL,
            custom_data=[C.TOOLTIP_COL],
            labels={
                C.X_AXIS_COL: simple_chart.x_axis_label,
                C.Y_AXIS_COL: simple_chart.y_axis_label,
                C.COLOR_COL: simple_chart.color_label,
            },
        )
    elif ct == M.SimpleChartType.STACKED_AREA:
        fig = px.area(
            pdf,
            x=C.X_AXIS_COL,
            y=C.Y_AXIS_COL,
            text=C.TEXT_COL,
            color=C.COLOR_COL,
            custom_data=[C.TOOLTIP_COL],
            labels={
                C.X_AXIS_COL: simple_chart.x_axis_label,
                C.Y_AXIS_COL: simple_chart.y_axis_label,
                C.COLOR_COL: simple_chart.color_label,
            },
        )
    elif ct == M.SimpleChartType.HEATMAP:
        args = pdf_to_heatmap(pdf)
        fig = ff.create_annotated_heatmap(
            colorscale="BuPu",
            **args,
        )
        fig = set_heatmap_figure_style(fig, simple_chart)
        return fig

    fig.update_traces(hovertemplate="%{customdata[0]} <extra></extra>")
    fig.update_layout(yaxis_ticksuffix=simple_chart.yaxis_ticksuffix)
    fig = set_figure_style(fig, simple_chart, metric)

    return fig


def figure_to_base64_image(
    figure, scale: float = 1.0, kaleid_configs: Tuple = None
) -> str:
    try:
        if kaleid_configs is not None:
            pio.kaleido.scope.mathjax = None
            pio.kaleido.scope.chromium_args += kaleid_configs

        img_bytes = figure.to_image(
            format="svg",
            scale=scale,
        )
        return "data:image/svg+xml;base64," + b64encode(img_bytes).decode()
    except BaseException:
        traceback.print_exc()

        return "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mO8XQ8AAjsBXM7pODsAAAAASUVORK5CYII="



from __future__ import annotations

import traceback
from typing import Any, Dict, Optional, cast
import dash.development.base_component as bc
import dash_bootstrap_components as dbc
import mitzu.model as M
import mitzu.webapp.pages.explore.toolbar_handler as TH
import mitzu.webapp.pages.explore.explore_page as EXP
import mitzu.webapp.configs as configs
import mitzu.webapp.dependencies as DEPS
import mitzu.webapp.storage as S
import mitzu.visualization.common as CO
import flask
import mitzu.webapp.pages.paths as P
import mitzu.webapp.model as WM
import mitzu.serialization as SE
from mitzu.webapp.helper import MITZU_LOCATION

import pandas as pd
from dash import Input, Output, State, ctx, dcc, html, callback, no_update, dash_table
from mitzu.webapp.helper import get_final_all_inputs
from mitzu.webapp.auth.decorator import restricted
import mitzu.visualization.plot as PLT
import mitzu.visualization.charts as CHRT
import json
import hashlib
from urllib.parse import urlparse

GRAPH = "graph"
MESSAGE = "lead fw-normal text-center h-100 w-100"
TABLE = "explore_results_table"
SQL_AREA = "sql_area"


CONTENT_STYLE = {
    "min-height": "200px",
    "max-height": "700px",
    "overflow": "auto",
}

DF_CACHE: Dict[int, pd.DataFrame] = {}
MARKDOWN = """```sql
{sql}
```"""

GRAPH_CONTAINER = (
    "graph_container d-flex justify-content-stretch align-items-center pt-3"
)
GRAPH_REFRESHER_INTERVAL = "graph_refresher_interval"


def create_graph_container() -> bc.Component:
    if configs.BACKGROUND_CALLBACK:
        return html.Div(
            id=GRAPH_CONTAINER,
            children=[],
            className=GRAPH_CONTAINER,
        )
    else:
        return html.Div(
            dcc.Loading(
                id=GRAPH_CONTAINER,
                children=[],
                color="#287bb5",
                type="dot",
                className=GRAPH_CONTAINER,
            ),
        )


def create_metric_hash_key(metric: M.Metric) -> str:
    metric_dict = SE.to_dict(metric)

    #  We need to remove these keys from the metric dict as changes in these shouldn't trigger reexecution
    metric_dict["co"]["mn"] = None
    metric_dict["co"]["mgc"] = None
    metric_dict["co"]["ct"] = None
    metric_dict["co"]["cat"] = None
    metric_dict["id"] = None

    return hashlib.md5(json.dumps(metric_dict).encode("ascii")).hexdigest()


def get_metric_result_df(
    hash_key: str, metric: M.Metric, storage: S.MitzuStorage
) -> pd.DataFrame:
    result_df = storage.get_query_result_dataframe(hash_key)
    if result_df is None:
        result_df = metric.get_df()
        storage.set_query_result_dataframe(hash_key, result_df)
    return result_df


def create_graph(
    metric: Optional[M.Metric],
    simple_chart: CO.SimpleChart,
) -> Optional[dcc.Graph]:
    if metric is None:
        return html.Div("Select the first event...", id=GRAPH, className=MESSAGE)

    if (
        (isinstance(metric, M.SegmentationMetric) and metric._segment is None)
        or (
            isinstance(metric, M.ConversionMetric)
            and len(metric._conversion._segments) == 0
        )
        or (isinstance(metric, M.RetentionMetric) and metric._initial_segment is None)
    ):
        return html.Div("Select the first event...", id=GRAPH, className=MESSAGE)

    fig = PLT.plot_chart(simple_chart, metric)
    return dcc.Graph(
        id=GRAPH, className="w-100", figure=fig, config={"displayModeBar": False}
    )


def create_table(
    metric: Optional[M.Metric], hash_key: str, storage: S.MitzuStorage
) -> Optional[dbc.Table]:
    if metric is None:
        return None

    result_df = get_metric_result_df(hash_key, metric, storage)
    result_df = result_df.sort_values(by=[result_df.columns[0], result_df.columns[1]])
    result_df.columns = [col[1:].replace("_", " ").title() for col in result_df.columns]
    table = dash_table.DataTable(
        id=TABLE,
        columns=[
            {"name": i, "id": i, "deletable": False, "selectable": False}
            for i in result_df.columns
        ],
        data=result_df.to_dict("records"),
        editable=False,
        sort_action="native",
        sort_mode="single",
        filter_action="native",
        export_columns="all",
        export_headers="names",
        export_format="csv",
        page_size=20,
        style_cell={"padding": "10px", "fontFamily": "var(--mdb-font-sans-serif)"},
        style_header={
            "backgroundColor": "var(--mdb-gray-100)",
            "fontWeight": "bold",
        },
    )

    return table


def create_sql_area(metric: Optional[M.Metric]) -> dbc.Table:
    if metric is not None:
        return dcc.Markdown(
            children=MARKDOWN.format(sql=metric.get_sql()),
            id=SQL_AREA,
        )
    else:
        return html.Div(id=SQL_AREA)


def store_rendered_saved_metric(
    metric_name: str,
    metric: M.Metric,
    simple_chart: CO.SimpleChart,
    project: M.Project,
    storage: S.MitzuStorage,
    metric_id: Optional[str] = None,
):
    fig = PLT.plot_chart(simple_chart, metric)
    small_image_b64 = PLT.figure_to_base64_image(
        fig, 0.5, kaleid_configs=configs.get_kaleido_configs()
    )
    saved_metric = WM.SavedMetric(
        metric=metric,
        chart=simple_chart,
        project=project,
        image_base64=small_image_b64,
        small_base64=small_image_b64,
        name=metric_name,
        id=metric_id,
    )

    storage.set_saved_metric(metric_id=saved_metric.id, saved_metric=saved_metric)


def create_callbacks():
    @callback(
        output=Output(GRAPH_CONTAINER, "children"),
        inputs=EXP.ALL_INPUT_COMPS,
        state=dict(
            graph_content_type=State(TH.GRAPH_CONTENT_TYPE, "value"),
            href=State(MITZU_LOCATION, "href"),
            metric_name=State(EXP.METRIC_NAME_INPUT, "value"),
            metric_id=State(EXP.METRIC_ID_VALUE, "children"),
        ),
        interval=configs.GRAPH_POLL_INTERVAL_MS,
        background=configs.BACKGROUND_CALLBACK,
        running=[
            (
                Output(TH.GRAPH_REFRESH_BUTTON, "disabled"),
                True,
                False,
            ),
            (
                Output(TH.CANCEL_BUTTON, "style"),
                TH.VISIBLE,
                TH.HIDDEN,
            ),
        ],
        cancel=[Input(TH.CANCEL_BUTTON, "n_clicks")],
    )
    @restricted
    def handle_changes_for_graph(
        all_inputs: Dict[str, Any],
        graph_content_type: str,
        href: str,
        metric_name: Optional[str],
        metric_id: str,
    ) -> bc.Component:
        try:
            parse_result = urlparse(href)
            project_id = P.get_path_value(
                P.PROJECTS_EXPLORE_PATH, parse_result.path, P.PROJECT_ID_PATH_PART
            )
            storage = cast(
                DEPS.Dependencies, flask.current_app.config.get(DEPS.CONFIG_KEY)
            ).storage
            project = storage.get_project(project_id)
            if project is None:
                return no_update
            all_inputs = get_final_all_inputs(all_inputs, ctx.inputs_list)
            all_inputs[EXP.METRIC_ID_VALUE] = metric_id
            all_inputs[EXP.METRIC_NAME_INPUT] = metric_name
            all_inputs[EXP.MITZU_LOCATION] = parse_result.query
            dp = project._discovered_project.get_value()
            if dp is None:
                return html.Div(
                    [
                        "Your project haven't been discovered yet",
                        dcc.Link(
                            f"Discover {project.project_name}",
                            href=P.create_path(
                                P.EVENTS_AND_PROPERTIES_PROJECT_PATH,
                                project_id=project.id,
                            ),
                        ),
                    ],
                    id=GRAPH,
                    className=MESSAGE,
                )

            metric, _, _ = EXP.create_metric_from_all_inputs(all_inputs, dp)

            if metric is None:
                return html.Div("Select an event", id=GRAPH, className=MESSAGE)

            should_save_metrics = all_inputs[EXP.METRIC_NAME_INPUT] is not None
            simple_chart: CO.SimpleChart

            hash_key = create_metric_hash_key(metric)
            if ctx.triggered_id in (TH.GRAPH_REFRESH_BUTTON, TH.GRAPH_RUN_QUERY_BUTTON):
                storage.clear_query_result_dataframe(hash_key)

            if (
                not project.webapp_settings.auto_refresh_enabled
                and ctx.triggered_id != TH.GRAPH_RUN_QUERY_BUTTON
                and storage.get_query_result_dataframe(hash_key) is None
            ):
                return html.Div(
                    "Click run to execute the query", id=GRAPH, className=MESSAGE
                )

            if graph_content_type == TH.CHART_VAL or should_save_metrics:
                result_df = get_metric_result_df(hash_key, metric, storage)
                simple_chart = CHRT.get_simple_chart(metric, result_df)

            if graph_content_type == TH.CHART_VAL:
                res = create_graph(metric, simple_chart)  # noqa
            if graph_content_type == TH.TABLE_VAL:
                res = create_table(metric, hash_key, storage)
            elif graph_content_type == TH.SQL_VAL:
                res = create_sql_area(metric)

            if should_save_metrics and metric_name is not None:
                store_rendered_saved_metric(
                    metric_name=metric_name,
                    metric=metric,
                    simple_chart=simple_chart,
                    project=project,
                    storage=storage,
                    metric_id=metric_id,
                )

            return res
        except Exception as exc:
            traceback.print_exc()
            return html.Div(
                [
                    html.B("Something has gone wrong. Details:"),
                    html.Pre(children=str(exc)),
                ],
                id=GRAPH,
                className="text-danger small",
            )
